from dataclasses import dataclass

from local_tuya.protocol import ProtocolConfig


@dataclass
class DeviceConfig:
    protocol: ProtocolConfig
    # Seconds to wait for more update commands in order to group them.
    debounce_updates: float = 0
    # Seconds to wait for state updates to confirm the command has been acknowledged.
    # 0 disables confirmation.
    confirm_timeout: float = 5



from local_tuya.domoticz.units.base import Unit, UnitCommand
from local_tuya.domoticz.units.ids import UnitId
from local_tuya.domoticz.units.manager import UnitManager
from local_tuya.domoticz.units.selector_switch import (
    SelectorSwitchStyle,
    selector_switch_unit,
)
from local_tuya.domoticz.units.set_point import set_point_unit
from local_tuya.domoticz.units.switch import switch_unit
from local_tuya.domoticz.units.temperature import temperature_unit
from local_tuya.domoticz.units.value_preprocessors import (
    compose,
    debounce,
    moving_average,
)



import json
import logging
from dataclasses import asdict, dataclass
from enum import IntEnum
from typing import Awaitable, Callable, Dict, Generic, Optional, TypeVar

try:
    import DomoticzEx
except ModuleNotFoundError:
    from local_tuya.domoticz.types import DomoticzEx

from local_tuya.domoticz.types import DomoticzUnit

logger = logging.getLogger(__name__)


class ColorMode(IntEnum):
    WHITE = 1
    TEMP = 2
    RGB = 3
    CUSTOM = 4


@dataclass
class Color:
    m: int
    t: int
    r: int
    g: int
    b: int
    cw: int
    ww: int


@dataclass
class UnitCommand:
    command: str
    level: float
    color_str: str

    def color(self) -> Optional[Color]:
        if self.command != "Set Color" or not self.color_str:
            return None
        return Color(**json.loads(self.color_str))


@dataclass
class UnitValues:
    n_value: int
    s_value: str
    color: Optional[Color] = None

    def color_str(self) -> str:
        if not self.color:
            return ""
        return json.dumps(asdict(self.color))


T = TypeVar("T")


class Unit(Generic[T]):
    def __init__(
        self,
        # ID of the unit in Domoticz.
        # It has to be unique for the hardware but not across hardware.
        id_: int,
        # Domoticz unit type.
        type_: str,
        # Name that will be suffixed to the hardware name.
        name: str,
        # Image ID, see `/json.htm?type=custom_light_icons` in Domoticz.
        image: int,
        # Convert the value of the unit into the Domoticz unit values.
        to_unit_values: Callable[[T], UnitValues],
        # Convert a Domoticz command into this units value.
        command_to_value: Optional[Callable[[UnitCommand], Optional[T]]] = None,
        # Function to call when a unit command is triggered.
        command_func: Optional[Callable[[T], Awaitable]] = None,
        # Options for unit creation.
        options: Optional[Dict[str, str]] = None,
    ):
        self.id = id_
        self._type = type_
        self._name = name
        self._image = image
        self._options = options
        self._unit: Optional[DomoticzUnit] = None
        self._to_unit_values = to_unit_values
        self._command_to_value = command_to_value
        self._command_func = command_func

    def ensure(self, unit: Optional[DomoticzUnit], device_name: str) -> None:
        if unit is None:
            logger.info("creating unit %s (%s)", self._name, self.id)
            full_name = f"{device_name} {self._name}"
            unit = DomoticzEx.Unit(
                Name=full_name,
                DeviceID=device_name,
                Unit=self.id,
                Image=self._image,
                TypeName=self._type,
                **({"Options": self._options} if self._options else {}),
            )
            unit.Create()
        else:
            # Update some unit attributes.
            unit.Image = self._image
            # Should be read/write but raises `AttributeError: readonly attribute`.
            # if self._options:
            #     unit.Options = self._options
            unit.Update(False)
        self._unit = unit

    async def on_command(self, command: UnitCommand) -> None:
        if self._command_to_value and self._command_func:
            value = self._command_to_value(command)
            if value is not None:
                await self._command_func(value)

    def update(self, value: T) -> None:
        if self._unit is None:
            raise RuntimeError(f"unit {self.id} {self._name} not registered")
        values = self._to_unit_values(value)
        update = False
        if values.n_value != self._unit.nValue:
            self._unit.nValue = values.n_value
            update = True
        if values.s_value != self._unit.sValue:
            self._unit.sValue = values.s_value
            update = True
        color = values.color_str()
        if color and color != self._unit.Color:
            self._unit.Color = color
            update = True
        if update:
            self._unit.Update(Log=True)



from typing import Awaitable, Callable

from local_tuya.domoticz.units.base import Unit, UnitCommand, UnitValues


def switch_unit(
    id_: int,
    name: str,
    image: int,
    command_func: Callable[[bool], Awaitable],
) -> Unit[bool]:
    def _to_unit_values(value: bool) -> UnitValues:
        if value:
            return UnitValues(n_value=1, s_value="On")
        return UnitValues(n_value=0, s_value="Off")

    def _command_to_value(command: UnitCommand) -> bool:
        return command.command.lower() == "on"

    return Unit(
        id_=id_,
        type_="Switch",
        name=name,
        image=image,
        to_unit_values=_to_unit_values,
        command_to_value=_command_to_value,
        command_func=command_func,
    )



import asyncio
import logging
from typing import Optional

from concurrent_tasks import BackgroundTask

from local_tuya.device.constraints import Constraints
from local_tuya.device.state import StateHandler
from local_tuya.protocol import Protocol, Values

logger = logging.getLogger(__name__)


class UpdateBuffer(BackgroundTask):
    """Debounce updates to the device."""

    def __init__(
        self,
        delay: float,
        confirm_timeout: float,
        protocol: Protocol,
        state_handler: StateHandler,
        constraints: Optional[Constraints],
    ):
        super().__init__(self._update)
        self._delay = delay
        self._confirm_timeout = confirm_timeout
        self._protocol = protocol
        self._state_handler = state_handler
        self._constraints = constraints
        self._buffer: Values = {}
        # Used to know we are currently performing an update.
        self._lock = asyncio.Lock()
        self._updated: Optional[asyncio.Future[None]] = None

    def __enter__(self) -> "UpdateBuffer":
        # Don't start the background task.
        return self

    async def update(self, values: Values) -> None:
        filtered = await self._filter({**self._buffer, **values})
        if not self._delay:
            if filtered:
                async with self._lock:
                    await self._do_update(filtered)
            return
        self._buffer = filtered
        async with self._lock:
            self.cancel()
            if self._buffer:
                self.create()
                if not self._updated:
                    self._updated = asyncio.Future()
            else:
                logger.debug(
                    "cancelling previous commands as update is no longer required"
                )
                if self._updated:
                    self._updated.set_result(None)
                    self._updated = None
        if self._updated:
            await self._updated

    async def _update(self) -> None:
        logger.debug(
            "received command waiting %ss before sending to device", self._delay
        )
        await asyncio.sleep(self._delay)
        buffer = self._buffer
        self._buffer = {}
        if buffer:
            async with self._lock:
                try:
                    await self._do_update(buffer)
                    if self._updated:
                        self._updated.set_result(None)
                except Exception as e:
                    if self._updated:
                        self._updated.set_exception(e)
                finally:
                    self._updated = None

    async def _do_update(self, values: Values) -> None:
        logger.info("updating device with: %s", values)
        await self._protocol.update(values)
        if not self._confirm_timeout:
            return
        # Wait until we receive real confirmation of update via status.
        await asyncio.wait_for(
            self._state_handler.matches(values),
            self._confirm_timeout,
        )

    async def _filter(self, values: Values) -> Values:
        state = await self._state_handler.state()
        filtered = {k: v for k, v in values.items() if state[k] != v}
        if not self._constraints:
            return filtered
        return self._constraints.filter_values(filtered, state)



"""Utilities to alter how device values are reported."""

import time
from typing import Callable, Optional, Tuple, TypeVar

T = TypeVar("T")

ValuePreprocessor = Callable[[T], T]


def compose(*preprocessors: ValuePreprocessor[T]) -> ValuePreprocessor[T]:
    """Compose multiple preprocessors.
    They are applied in the order in which they are passed.

    Ex:
    >>> compose(moving_average(5), debounce(30))
    """

    def _wrapper(value: T) -> T:
        for preprocessor in preprocessors:
            value = preprocessor(value)
        return value

    return _wrapper


def moving_average(n: int) -> ValuePreprocessor[float]:
    """Return the moving average over the last n values."""
    values: Tuple[float, ...] = ()

    def _wrapper(value: float) -> float:
        nonlocal values
        values = (*values[1 - n :], value)
        return sum(values) / len(values)

    return _wrapper


def debounce(s: float) -> ValuePreprocessor[float]:
    """Debounce the updates and return the first value within a period of s seconds."""
    last: Optional[float] = None
    last_time: float = 0

    def _wrapper(value: float) -> float:
        nonlocal last, last_time
        now = time.monotonic()
        if last is None or now >= last_time + s:
            last = value
            last_time = now
        return last

    return _wrapper



import asyncio
import logging
from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import Dict, Optional, Tuple, Type

from concurrent_tasks import RestartableTask

from local_tuya.errors import CommandTimeoutError
from local_tuya.events import EventNotifier
from local_tuya.protocol.events import (
    CommandSent,
    ConnectionEstablished,
    ConnectionLost,
    DataSent,
    ResponseReceived,
)
from local_tuya.protocol.message import Command, HeartbeatCommand, MessageHandler

logger = logging.getLogger(__name__)


class Sender(AbstractAsyncContextManager):
    def __init__(
        self,
        message_handler: MessageHandler,
        event_notifier: EventNotifier,
        timeout: float,
    ):
        event_notifier.register(ConnectionEstablished, self._connection_established)
        event_notifier.register(ConnectionLost, self._connection_lost)
        event_notifier.register(ResponseReceived, self._receive_response)
        event_notifier.register(CommandSent, self._send)
        self._notifier = event_notifier
        self._handler = message_handler
        self._timeout = timeout
        self._pending_tasks: Dict[
            Tuple[int, Optional[Type[Command]]], RestartableTask[None]
        ] = {}
        self._can_send = False
        self._sequence_number = 0

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        # This has to be async to let task cancellation be effective.
        for task in self._pending_tasks.values():
            task.cancel()

    def _connection_established(self, _: ConnectionEstablished) -> None:
        self._can_send = True
        for task in self._pending_tasks.values():
            task.start()

    def _connection_lost(self, _: ConnectionLost) -> None:
        self._can_send = False
        for task in self._pending_tasks.values():
            task.cancel()

    def _receive_response(self, event: ResponseReceived) -> None:
        task = self._pending_tasks.pop(
            (event.sequence_number, event.command_class), None
        )
        if task:
            if event.response.error:
                task.set_exception(event.response.error)
            else:
                task.set_result(None)

    async def _send(self, event: CommandSent) -> None:
        """Send a message, waiting for the response.
        Timeout is for waiting the response, it will wait until a connection can be acquired.
        """
        sequence_number = self._get_sequence_number(event.command)
        logger.debug(
            "sending message %i %s", sequence_number, event.command.__class__.__name__
        )
        data = self._handler.pack(sequence_number, event.command)
        task = RestartableTask[None](
            partial(self._notifier.emit, DataSent(data)),
            timeout=self._timeout,
        )
        if self._can_send:
            task.start()
        self._pending_tasks[(sequence_number, type(event.command))] = task
        try:
            await task
        except asyncio.TimeoutError as e:
            raise CommandTimeoutError() from e

    def _get_sequence_number(self, command: Command) -> int:
        if isinstance(command, HeartbeatCommand):
            # Devices return 0 whatever is sent.
            return 0
        if self._sequence_number == 1000:
            self._sequence_number = 1
        else:
            self._sequence_number += 1
        return self._sequence_number



from logging import DEBUG, INFO, WARNING, Formatter, LogRecord, StreamHandler

try:
    import DomoticzEx
except ModuleNotFoundError:
    from local_tuya.domoticz.types import DomoticzEx


class DomoticzHandler(StreamHandler):
    """Log handler to send to Domoticz logger."""

    def __init__(self):
        super().__init__()
        self.setFormatter(Formatter("%(levelname)s: %(name)s: %(message)s"))

    def emit(self, record: LogRecord) -> None:
        msg = self.format(record)
        if record.levelno < DEBUG:
            DomoticzEx.Debug(msg)
        elif record.levelno < INFO:
            DomoticzEx.Log(msg)
        elif record.levelno < WARNING:
            DomoticzEx.Status(msg)
        else:
            DomoticzEx.Error(msg)



from typing import List, Type

from local_tuya.protocol.config import ProtocolConfig
from local_tuya.protocol.message.handlers.handler import MessageHandler
from local_tuya.protocol.message.handlers.v33 import V33MessageHandler

HANDLERS: List[Type[MessageHandler]] = [
    V33MessageHandler,
]


def get_handler(config: ProtocolConfig) -> MessageHandler:
    for handler_class in HANDLERS:
        handler_ = handler_class.from_config(config)
        if handler_:
            return handler_
    raise ValueError(f"unsupported device: {config!r}")



from __future__ import annotations

import sys
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type, Union

import xmltodict

from local_tuya.domoticz.units import UnitId
from local_tuya.protocol import Version

if sys.version_info < (3, 8):
    from importlib_metadata import metadata as pkg_metadata
else:
    from importlib.metadata import metadata as pkg_metadata

XML = Union[str, Dict[str, "XML"], List["XML"]]


def _filter_xml_dict(d: Dict[str, XML]) -> Dict[str, XML]:
    return {k: v for k, v in d.items() if v}


def _xml_bool(b: bool) -> str:
    return "true" if b else ""


@dataclass
class Option:
    label: str
    value: str
    default: bool = False

    def to_xml_dict(self) -> XML:
        return _filter_xml_dict(
            {
                "@label": self.label,
                "@value": self.value,
                "@default": _xml_bool(self.default),
            }
        )


@dataclass
class Parameter:
    field: str
    label: str
    required: bool = False
    default: str = ""
    password: bool = False
    description: XML = ""
    options: Tuple[Option, ...] = ()

    def to_xml_dict(self) -> XML:
        return _filter_xml_dict(
            {
                "@field": self.field,
                "@label": self.label,
                "@required": _xml_bool(self.required),
                "@default": self.default,
                "@password": _xml_bool(self.password),
                "description": self.description,
                "options": _filter_xml_dict(
                    {"option": [o.to_xml_dict() for o in self.options]}
                ),
            }
        )


def get_package_metadata(package: str) -> Mapping[str, str]:
    return pkg_metadata(package)  # type: ignore


@dataclass
class PluginMetadata:
    # Name that will appear in the hardware dropdown list.
    name: str
    # Pypi package to pull author, version data.
    # Will also be used for the location of the plugin in the `plugins` directory.
    package: str
    # If not a string, should be compatible with `xmltodict.unparse`.
    # See https://github.com/martinblech/xmltodict#roundtripping.
    description: XML
    wiki_link: str = ""
    external_link: str = ""
    parameters: Tuple[Parameter, ...] = ()

    def definition(self, unit_ids: Optional[Type[UnitId]]) -> str:
        package_metadata = get_package_metadata(self.package)
        xml = {
            "plugin": _filter_xml_dict(
                {
                    "@key": self.package,
                    "@name": self.name,
                    "@author": package_metadata.get("Author", "unknown"),
                    "@version": package_metadata.get("Version", "1.0.0"),
                    "@wikilink": self.wiki_link,
                    "@externallink": self.external_link
                    or package_metadata.get("Home-page", ""),
                    "description": self.description,
                    "params": {
                        "param": [
                            p.to_xml_dict()
                            for p in (
                                Parameter(
                                    field="Username", label="Device ID", required=True
                                ),
                                Parameter(
                                    field="Address",
                                    label="Device IP address",
                                    required=True,
                                ),
                                Parameter(
                                    field="Port", label="Device port", default="6668"
                                ),
                                Parameter(
                                    field="Password",
                                    label="Device local key",
                                    required=True,
                                    password=True,
                                ),
                                *self.parameters,
                                Parameter(
                                    field="Mode4",
                                    label="Tuya version",
                                    options=tuple(
                                        Option(
                                            label=v.decode(),
                                            value=v.decode(),
                                            default=v == Version.v33,
                                        )
                                        for v in Version
                                    ),
                                ),
                                *(
                                    (
                                        Parameter(
                                            field="Mode5",
                                            label="Included units",
                                            default=unit_ids.names(),
                                        ),
                                    )
                                    if unit_ids
                                    else ()
                                ),
                                Parameter(
                                    field="Mode6",
                                    label="Debug",
                                    options=(
                                        Option(label="No", value="0", default=True),
                                        Option(label="Yes", value="1"),
                                    ),
                                ),
                            )
                        ],
                    },
                }
            ),
        }
        return xmltodict.unparse(
            xml,
            full_document=False,
            pretty=True,
            short_empty_elements=True,
            indent="  ",
        )



from local_tuya.device.config import DeviceConfig
from local_tuya.device.constraints import Constraint, Constraints
from local_tuya.device.device import Device
from local_tuya.device.enums import DataPoint, State



import asyncio
import logging
from typing import Optional

from concurrent_tasks import BackgroundTask

from local_tuya.errors import CommandTimeoutError
from local_tuya.events import EventNotifier
from local_tuya.protocol.events import (
    CommandSent,
    ConnectionEstablished,
    ConnectionLost,
    ResponseReceived,
    StateUpdated,
)
from local_tuya.protocol.message import (
    StateCommand,
    StateResponse,
    StatusResponse,
    Values,
)

logger = logging.getLogger(__name__)


class State(BackgroundTask):
    def __init__(
        self,
        refresh_interval: float,
        event_notifier: EventNotifier,
    ):
        super().__init__(self._refresh, refresh_interval)
        event_notifier.register(ResponseReceived, self._update)
        event_notifier.register(ConnectionLost, lambda _: self.cancel())
        event_notifier.register(ConnectionEstablished, lambda _: self.create())
        self._notifier = event_notifier
        self._state: Optional[Values] = None

    def __enter__(self) -> "State":
        return self

    async def _refresh(self, interval: float) -> None:
        while True:
            logger.debug("refreshing device state")
            try:
                await self._notifier.emit(CommandSent(StateCommand()))
            except CommandTimeoutError:
                logger.warning("timeout waiting for state response")
            await asyncio.sleep(interval)

    async def _update(self, event: ResponseReceived) -> None:
        if event.response.error:
            pass
        elif isinstance(event.response, StateResponse):
            new_state = event.response.values
            if self._state != new_state:
                logger.debug("received new device state: %s", new_state)
                self._state = new_state
                await self._notifier.emit(StateUpdated(self._state))
        elif isinstance(event.response, StatusResponse):
            if self._state is None:
                # We have not yet received the initial state.
                # Better discard this as the state might be newer.
                return
            new_state = {**self._state, **event.response.values}
            if self._state != new_state:
                logger.debug("received device state update: %s", event.response.values)
                self._state = new_state
                await self._notifier.emit(StateUpdated(self._state))



from enum import IntEnum
from typing import Optional, Set


class UnitId(IntEnum):
    @classmethod
    def names(cls) -> str:
        return ",".join(e.name.lower() for e in cls)

    @classmethod
    def included(cls, names: Optional[str]) -> Set["UnitId"]:
        if not names:
            return set(cls)
        included_names = names.split(",")
        return {e for e in cls if e.name.lower() in included_names}



import logging

from local_tuya.backoff import Backoff, SequenceBackoff
from local_tuya.device import (
    Constraint,
    Constraints,
    DataPoint,
    Device,
    DeviceConfig,
    State,
)
from local_tuya.errors import (
    CommandTimeoutError,
    DecodeResponseError,
    LocalTuyaError,
    ResponseError,
)
from local_tuya.protocol import ProtocolConfig, Values

logging.getLogger(__name__).addHandler(logging.NullHandler())



from enum import Enum
from typing import Awaitable, Callable, Dict, List, Optional, Type, TypeVar

from local_tuya.domoticz.units.base import Unit, UnitCommand, UnitValues

T = TypeVar("T", bound=Enum)


class SelectorSwitchStyle(str, Enum):
    BUTTON_BAR = "0"
    DROP_DOWN = "1"


def selector_switch_unit(
    id_: int,
    name: str,
    image: int,
    # All possible values.
    # Names will be used for labels.
    # Values should correspond to what is sent to the tuya device.
    enum: Type[T],
    command_func: Callable[[T], Awaitable],
    off_label: str = "Off",
    off_hidden: bool = True,
    style: SelectorSwitchStyle = SelectorSwitchStyle.BUTTON_BAR,
) -> Unit[T]:
    _value_to_level: Dict[T, int] = {}
    _level_to_value: Dict[int, T] = {}
    labels: List[str] = [off_label]
    e: T
    for i, e in enumerate(enum, start=1):
        _value_to_level[e] = i * 10
        _level_to_value[i * 10] = e
        labels.append(f"{e.name[0]}{e.name[1:].lower()}")  # type: ignore

    def _to_unit_values(value: T) -> UnitValues:
        level = _value_to_level.get(value, 0)
        return UnitValues(
            n_value=1 if level else 0,
            s_value=str(level),
        )

    def _command_to_value(command: UnitCommand) -> Optional[T]:
        return _level_to_value.get(int(command.level))

    return Unit(
        id_=id_,
        type_="Selector Switch",
        name=name,
        image=image,
        to_unit_values=_to_unit_values,
        command_to_value=_command_to_value,
        command_func=command_func,
        options={
            "LevelActions": "|".join(["" for _ in labels]),
            "LevelNames": "|".join(labels),
            "LevelOffHidden": str(off_hidden).lower(),
            "SelectorStyle": style.value,
        },
    )



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Type

from local_tuya.protocol.config import ProtocolConfig
from local_tuya.protocol.message.messages import Command, Response


class MessageHandler(ABC):
    @staticmethod
    @abstractmethod
    def from_config(config: ProtocolConfig) -> Optional["MessageHandler"]:
        """Create the handler if it supports the device."""

    @abstractmethod
    def pack(self, sequence_number: int, message: Command) -> bytes:
        """Pack a message into bytes."""

    @abstractmethod
    def unpack(
        self, data: bytes
    ) -> Tuple[int, Optional[Response], Optional[Type[Command]], bytes]:
        """Extract message from the bytes received, also return unused remaining bytes."""



from abc import ABC, abstractmethod
from enum import Enum

from local_tuya.protocol import Values


class DataPoint(str, Enum):
    """Data points supported by the device."""


class State(ABC):
    """Typed and user-friendly wrapper around state of the device.
    Prefer using dataclasses or anything generating a proper __repr__ method for debugging.

    Usage:
    >>> from dataclasses import dataclass
    >>>
    >>> @dataclass
    >>> class SwitchState(State):
    >>>     power: bool
    >>>
    >>>     @classmethod
    >>>     def load(cls, values: Values) -> "SwitchState":
    >>>         return cls(power=bool(values[SwitchDataPoint.POWER]))
    """

    @classmethod
    @abstractmethod
    def load(cls, values: Values) -> "State":
        """Create from raw values."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['local_tuya',
 'local_tuya.device',
 'local_tuya.domoticz',
 'local_tuya.domoticz.plugin',
 'local_tuya.domoticz.units',
 'local_tuya.protocol',
 'local_tuya.protocol.message',
 'local_tuya.protocol.message.handlers']

package_data = \
{'': ['*']}

install_requires = \
['concurrent-tasks>=1.3,<2', 'pycryptodomex>=3,<4', 'xmltodict>=0.13,<0.14']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4,<5']}

setup_kwargs = {
    'name': 'local-tuya',
    'version': '1.2.2',
    'description': 'Interface to Tuya devices over LAN.',
    'long_description': "# local-tuya\n\n[![tests](https://github.com/gpajot/local-tuya/workflows/Test/badge.svg?branch=main&event=push)](https://github.com/gpajot/local-tuya/actions?query=workflow%3ATest+branch%3Amain+event%3Apush)\n[![version](https://img.shields.io/pypi/v/local-tuya?label=stable)](https://pypi.org/project/local-tuya/)\n[![python](https://img.shields.io/pypi/pyversions/local-tuya)](https://pypi.org/project/local-tuya/)\n\nInterface to Tuya devices over LAN.\n\nFeatures:\n- asynchronous methods and transport\n- persistent communication to the device\n- automatic remote device state updates (remotes can still be used)\n- configuratble of buffering for subsequent updates\n- constraints between device commands\n- Domoticz plugin using a dedicated thread\n\n> 💡 For now, only v3.3 is supported as I only own devices using this version.\n\n## Requirements\nTo control a device you will need these 3 things:\n- the device ID\n- the device local IP address\n- the device local key (encryption key generated upon pairing)\n\n> ⚠️ This library does not provide support for getting these.\n> See how to do that using any of those projects:\n> - [tuyapi](https://github.com/codetheweb/tuyapi)\n> - [tinytuya](https://github.com/jasonacox/tinytuya)\n> \n> Generous thanks to the maintainers of those tools for details on interfacing with Tuya devices.\n\n> ⚠️ Keep in mind that:\n> - After pairing the devices, it's recommended to assign static IPs in your router.\n> - If you reset or re-pair devices the local key will change.\n> - You can delete your tuya IOT account but not the SmartLife one and devices should be kept there.\n> - For state updates to be received properly, the device needs to be able to access the Tuya backend.\n\n## Architecture\nThis library is composed of two main components:\n- the Tuya protocol\n- the device\n\n### Protocol\nThe protocol is responsible for handling communication details with the Tuya device.\nIts interface consists of an asynchronous method to update the device and accepts a callback to subscribe to state changes.\n\nSee [protocol module](./local_tuya/protocol).\n\n### Device\nThe device handles higher level functional logic such as buffering, constraints and specific device commands.\n\nSee [device module](./local_tuya/device).\n\n## Domoticz plugin tools\nSee [Domoticz module](./local_tuya/domoticz).\n",
    'author': 'Gabriel Pajot',
    'author_email': 'gab@les-cactus.co',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpajot/local-tuya',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)



import inspect
from abc import ABC
from collections import defaultdict
from typing import Any, Awaitable, Callable, DefaultDict, List, Type, TypeVar, cast

from typing_extensions import ParamSpec


class Event(ABC):  # noqa: B024
    """Base event class"""


T = TypeVar("T", bound=Event)


class EventNotifier:
    def __init__(self):
        self._listeners: DefaultDict[
            Type[Event], List[Callable[[Event], Awaitable[None]]]
        ] = defaultdict(list)

    def register(self, event_class: Type[T], listener: Callable[[T], Any]) -> None:
        self._listeners[event_class].append(
            cast(
                Callable[[Event], Awaitable[None]],
                maybe_async(listener),
            )
        )

    async def emit(self, event: Event) -> None:
        for listener in self._listeners[type(event)]:
            await listener(event)


P = ParamSpec("P")


def maybe_async(func: Callable[P, Any]) -> Callable[P, Awaitable[None]]:
    async def _wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
        res = func(*args, **kwargs)
        if inspect.iscoroutine(res):
            await res

    return _wrapper



import logging
from typing import Callable, Dict, Generic, Optional, Type, TypeVar

from concurrent_tasks import BlockingThreadedTaskPool, ThreadedPoolContextManagerWrapper

from local_tuya.device import Device, State
from local_tuya.domoticz.logger import DomoticzHandler
from local_tuya.domoticz.types import DomoticzDevice
from local_tuya.domoticz.units import UnitCommand, UnitId, UnitManager
from local_tuya.protocol import ProtocolConfig, Version

try:
    import DomoticzEx
except ModuleNotFoundError:
    from local_tuya.domoticz.types import DomoticzEx

logger = logging.getLogger(__name__)
LOG_HANDLER = DomoticzHandler()

T = TypeVar("T", bound=State)
OnStart = Callable[[ProtocolConfig, Dict[str, str], UnitManager[T]], Device[T]]


class Plugin(Generic[T]):
    """Defines the Domoticz plugin.
    It connects the device to Domoticz units.
    """

    def __init__(
        self,
        package: str,
        on_start: OnStart,
        unit_ids: Optional[Type[UnitId]],
    ):
        self._package = package
        self._on_start = on_start
        self._unit_ids = unit_ids
        self._manager: Optional[UnitManager[T]] = None
        self._task_pool: Optional[BlockingThreadedTaskPool] = None
        # Setup loggers to log in Domoticz.
        for pkg in ("local_tuya", package):
            _logger = logging.getLogger(pkg)
            _logger.addHandler(LOG_HANDLER)

    @staticmethod
    def _protocol_config(parameters: Dict[str, str]) -> ProtocolConfig:
        return ProtocolConfig(
            id_=parameters["Username"],
            address=parameters["Address"],
            port=int(parameters["Port"]),
            key=parameters["Password"].encode(),
            version=Version(parameters["Mode4"].encode()),
        )

    def _setup_logging(self, debug: bool) -> None:
        DomoticzEx.Debugging(2 + 4 + 8 if debug else 0)
        # Setup loggers to log in Domoticz.
        for pkg in ("local_tuya", self._package):
            _logger = logging.getLogger(pkg)
            _logger.setLevel(logging.DEBUG if debug else logging.INFO)

    def start(
        self,
        parameters: Dict[str, str],
        devices: Dict[str, DomoticzDevice],
    ) -> None:
        """Start the device in a separate thread."""
        self.stop()
        self._setup_logging(parameters.get("Mode6", "") == "1")
        DomoticzEx.Heartbeat(15)
        name = parameters["Name"]
        included_units = parameters.get("Mode5", "")
        manager: UnitManager[T] = UnitManager(
            name=name,
            units=devices[name].Units if name in devices else {},
            included_units=self._unit_ids.included(included_units)
            if self._unit_ids and included_units
            else None,
        )
        self._manager = manager

        def _get_device() -> Device[T]:
            device = self._on_start(
                self._protocol_config(parameters), parameters, manager
            )
            device.set_state_updated_callback(manager.update)
            return device

        self._task_pool = BlockingThreadedTaskPool(
            context_manager=ThreadedPoolContextManagerWrapper(_get_device),
        )
        self._task_pool.start()

    def stop(self) -> None:
        """Stop the device if started."""
        if self._task_pool:
            self._task_pool.stop()
            self._task_pool = None
        if self._manager:
            self._manager.cleanup_domoticz_references()
            self._manager = None

    def on_command(self, unit_id: int, command: UnitCommand) -> None:
        """Send an command to the device asynchronously."""
        if self._task_pool and self._manager:
            # Fire and forget.
            self._task_pool.create_task(
                self._on_command(self._manager, unit_id, command)
            )
        else:
            logger.warning(
                "error sending command %s to unit %d, plugin %s not started",
                command,
                unit_id,
                self.__class__.__qualname__,
            )

    async def _on_command(
        self,
        manager: UnitManager[T],
        unit_id: int,
        command: UnitCommand,
    ) -> None:
        """Suppresses and logs exceptions."""
        try:
            await manager.on_command(unit_id, command)
        except Exception:
            logger.exception(
                "error sending command %s to unit %s of plugin %s",
                command,
                unit_id,
                self.__class__.__qualname__,
            )
